from aw.keyword_operations_base import KeywordOperationsBase

__author__ = 'mattjmorris'

import json
import daf.get as daf_get
import logging
from time import sleep


class KeywordOperationsMS(KeywordOperationsBase):

    def modify_bids(self, bid_modifier_func=lambda x: x, kw_df=None, adgroup_ids=None):
        """
        In dataframe format, gets all of the keywords in the current account or set of ad groups within the current
        account (if a list of ad groups is passed in), modifies all of the bids by using the bid_modifier_func, stores
        the resulting dataframes in local storage (for purposes of tracking), and pushes the modifications up to
        Google via bulk mutate.

        Inputs:
        bid_modifier_func: bid modification function, should operate on just the bid value.
        kw_df = pass in a dataframe of the keywords you want modified, or none to force a lookup. If you pass it in,
          must have column names that match those expected by the Operations class.
        adgroup_ids: If you pass in your own kw_df, leave this as None. Otherwise, pass in a list of adgroup ids, or
          none to modify the entire account
        """
        # We reset here so we don't accidentally append to old jobs
        self.job_ids = []

        kw_df = kw_df or self.awq.get_positive_keywords(adgroup_ids)
        # sorting the keywords by adgroup helps speed things up when we later push changes to google
        kw_df.sort(columns=['ad_group_id'], inplace=True)

        kw_df.max_cpc = kw_df.max_cpc.apply(bid_modifier_func)

        kw_df['ops'] = kw_df.apply(self.ops.update_keyword_df_row, axis=1)

        big_one = kw_df > self.chk_size * 5
        first_push = True

        self.job_ids = []

        # chunk into appropriate sizes, to constrain number of ops in each job
        for df in daf_get.chunks(kw_df, self.chk_size):

            # Google requests that we wait 2 seconds between multiple pushes to the same account
            if big_one and not first_push:
                sleep(2)
            first_push = False

            job_id = self.mutations.modify_keywords(df.ops.apply(json.loads).tolist())
            self.job_ids.append(job_id)

            # Replace any negative signs with an 'N' so that HDF5 Store stops giving warnings.
            self.store.put('jobs/_' + job_id.replace('-', 'N'), df)
            self.logger.log(level=logging.DEBUG, ko="Pushed new job to google, id = {}".format(job_id))

        return self.job_ids


__author__ = 'thorwhalen'
"""
Includes functions to diagnose duplicates in adwords elements
"""

from ut.aw.manip import add_col
from ut.daf.dup_diag import get_duplicates
import ut.coll.order_conserving as oc
from ut.aw.manip import assert_dependencies
import pandas as pd
import ut.daf.ch as daf_ch
import ut.daf.dup_diag as daf_dup_diag
import ut.aw.manip as aw_manip
import ut.pstr.trans as pstr_trans
import numpy as np
import collections
from ut.pstr.trans import to_unicode_or_bust

def general_stats_for_dup_diag(diag_df, n_taps=None, n_broad_taps=None, dup_types=['strip','lower','ascii']):
    if not _is_a_kw_dup_diagnosis_df(diag_df):
        # assume it is the original dataframe, so perform kw_dup_diagnosis on it (and compute n_taps and n_broad taps
        n_taps = len(diag_df)
        n_broad_taps = len(diag_df)
        diag_df = kw_dup_diagnosis(diag_df)
        return_diag_df = True
    else:
        return_diag_df = False
    # initialize stats dict
    stats = collections.OrderedDict()
    # compute the any_dup stats
    stats['number_of_kws'] = n_taps
    stats['any_dup_kw_count'] = len(diag_df)
    stats['any_dup_kw_ratio'] = len(diag_df)/float(n_taps)
    # compute the dup type stats
    for dup_type in dup_types:
        stats[dup_type+'_dup_kw_count'] = _total_dups(diag_df['dups_'+dup_type])
        stats[dup_type+'_dup_kw_ratio'] = stats[dup_type+'_dup_kw_count']/float(n_taps)
        stats[dup_type+'_dup_group_count'] = _total_dup_groups(diag_df['grp_id_'+dup_type])
    if 'dups_order' in diag_df.columns:
        dup_type = 'order'
        d = diag_df[diag_df.match_type=='Broad']
        stats[dup_type+'_dup_kw_count'] = _total_dups(d['dups_'+dup_type])
        if n_broad_taps!=0:
            stats[dup_type+'_dup_kw_ratio'] = stats[dup_type+'_dup_kw_count']/float(n_broad_taps)
        else:
            stats[dup_type+'_dup_kw_ratio'] = 0
        stats[dup_type+'_dup_group_count'] = _total_dup_groups(d['grp_id_'+dup_type])
    # return the stats dict, and also the diag_df if it wasn't given in thepowerinput
    if return_diag_df==True:
        named_tuple = collections.namedtuple('dup_stats',['stats','diag_df'])
        return named_tuple(stats=stats,diag_df=diag_df)
    else:
        return stats

def _is_a_kw_dup_diagnosis_df(df):
    if 'dups_strip' in df.columns:
        return True
    else:
        return False

def _total_dups(array_of_dup_counts):
    return len([x for x in array_of_dup_counts if x!=0])

def _total_dup_groups(array_of_dup_counts):
    return len(np.unique([x for x in array_of_dup_counts if x!=0]))

def _grp_id_array_to_number_of_groups(grp_id_array):
    unik_grp_id_array = np.unique(grp_id_array)
    nDupGroups = len(unik_grp_id_array)
    if isinstance(unik_grp_id_array[0],str):
        if '' in unik_grp_id_array:
            nDupGroups = nDupGroups - 1
    else:
        if 0 in unik_grp_id_array:
            nDupGroups = nDupGroups - 1
    return nDupGroups

def kw_dup_diagnosis(df, grp_keys=['match_type'], # grp_keys=['match_type','ad_group','campaign']?
                     grp_fun_dict={'dups': lambda x: len(x)}, grp_id_name='grp_id',grp_id_type='int',output_nondup_df=False):
    dup_df_dict = dict()
    grp_keys = oc.intersect(df.columns, grp_keys) + ['kw_representative']
    df = df.copy() # to change the input df (can be handled differently if need to spare memory)
    df.keyword = df.keyword.apply(lambda x:to_unicode_or_bust(x)) # change all keyword strings to unicode

    # util function (returns a dataframe containing grp_id and dups of a df
    def _get_grp_id_and_dups(df):
        """
        this function makes grp_id and dups duplication info columns and returns only those rows with dups>1
        NOTE: It is not meant to be used externally, but by the kw_dup_diagnosis() only
        """
        df = daf_dup_diag.ad_group_info_cols(df, grp_keys=grp_keys,
                                            grp_fun_dict=grp_fun_dict,
                                            grp_id_name=grp_id_name,
                                            grp_id_type=grp_id_type
        )
        if len(df)>0:
            return df[['grp_id', 'dups']][df.dups>1]
        else: # return an empty dataframe (but with the usual columns (necessary for the further joins)
            return pd.DataFrame(columns=['grp_id', 'dups'])

    # make a kw_representative column where different "group representatives" will be placed
    df['kw_representative'] = df['keyword']
    # get the kw_stripped duplicates
    df['kw_representative'] = aw_manip.strip_kw(df['kw_representative'])
    dup_df_dict['strip'] = _get_grp_id_and_dups(df)
    # get the kw_lower duplicates
    df['kw_representative'] = pstr_trans.lower(df['kw_representative'])
    dup_df_dict['lower'] = _get_grp_id_and_dups(df)
    # get the ascii duplicates
    df['kw_representative'] = pstr_trans.toascii(df['kw_representative'])
    dup_df_dict['ascii'] = _get_grp_id_and_dups(df)
    # get the order duplicates (only for Broads)
    d = df[df.match_type=='Broad']
    d['kw_representative'] = aw_manip.order_words(d['kw_representative'])
    dup_df_dict['order'] = _get_grp_id_and_dups(d)
    # join all this together
    d = dup_df_dict['strip'].join(dup_df_dict['lower'],how='outer',lsuffix='_strip').fillna(0)
    d = d.join(dup_df_dict['ascii'],how='outer',lsuffix='_lower').fillna(0)
    d = d.join(dup_df_dict['order'],how='outer',lsuffix='_ascii',rsuffix='_order').fillna(0)
    del df['kw_representative']
    d = d.join(df)
    if output_nondup_df==False:
        return d
    else:
        named_tuple = collections.namedtuple('dup_stats',['dup_diag_df','non_dup_df'])
        return named_tuple(dup_diag_df=d, non_dup_df=df.ix[list(set(df.index)-set(d.index))])

# def _get_grp_id_and_dups(df, grp_keys, grp_fun_dict, grp_id_name):
#     """
#     this function makes grp_id and dups duplication info columns and returns only those rows with dups>1
#     NOTE: It is not meant to be used externally, but by the kw_dup_diagnosis() only
#     """
#     df = daf_dup_diag.ad_group_info_cols(df, grp_keys=grp_keys,
#                                         grp_fun_dict=grp_fun_dict,
#                                         grp_id_name=grp_id_name
#     )
#     return df[['grp_id','dups']][df.dups>1]

def get_kw_duplicates_01(df,dup_def='kw_lower',gr_keys=['match_type','ad_group','campaign']):
    """
    old function to get kw_duplicates
    probably better to use kw_dup_diagnosis
    """
    if dup_def=='all':
        # get kw_lower dup_count df
        d = get_kw_duplicates_01(df,dup_def='kw_lower',gr_keys=gr_keys)
        d = daf_ch.ch_col_names(d,'dup_count','lower_dups')
        del d['kw_lower']
        # get kw_lower_ascii dup_count df
        dd = get_kw_duplicates_01(df,dup_def='kw_lower_ascii',gr_keys=gr_keys)
        dd = daf_ch.ch_col_names(dd,'dup_count','ascii_dups')
        del dd['kw_lower_ascii']
        # merge d and dd
        d = d.merge(dd,how='outer')
        # get kw_lower_ascii_ordered dup_count df
        dd = get_kw_duplicates_01(df,dup_def='kw_lower_ascii_ordered',gr_keys=gr_keys)
        dd = daf_ch.ch_col_names(dd,'dup_count','order_dups')
        del dd['kw_lower_ascii_ordered']
        # merge d and dd
        d = d.merge(dd,how='outer')
        # replace nans with 0s
        d = d.fillna(0)
        return d
    else:
        df = df.copy() # make a copy
        if dup_def=='kw_lower':
            d = add_col(df,'kw_lower',overwrite=False)
            gr_keys = oc.union(['kw_lower'],gr_keys)
        elif dup_def=='kw_lower_ascii':
            d = add_col(df,'kw_lower_ascii',overwrite=False)
            gr_keys = oc.union(['kw_lower_ascii'],gr_keys)
        elif dup_def=='kw_lower_ascii_ordered':
            d = df[df.match_type=='Broad']
            d = add_col(d,'kw_lower_ascii_ordered',overwrite=False)
            gr_keys = oc.union((['kw_lower_ascii_ordered'],gr_keys))
        else:
            raise ValueError("don't know how to handle that dup_def")
        assert_dependencies(d,gr_keys,"to get duplicates")
        return get_duplicates(d,gr_keys,keep_count=True)


def group_normalized_freq(arr):
    """
    transformation: value divided by the sum of values in the array
    """
    return arr/float(sum(arr))

def group_normalized_count(arr):
    """
    aggregation: inverse of array length
    """
    return 1.0/float(len(arr))


# def get_duplicates(df,cols):
#     df = df.reindex(index=range(len(df)))
#     grouped = df.groupby(cols)
#     unique_index = [gp_keys[0] for gp_keys in grouped.groups.values()]
#     non_unique_index = list(set(df.index)-set(unique_index))
#     duplicates_df = get_unique(df.irow(non_unique_index),cols)
#     duplicates_df = duplicates_df[cols]
#     return df.merge(pd.DataFrame(duplicates_df))


# if __name__=="__main__":
#     print "loading some data"
#     df = pd.load('/D/Dropbox/dev/py/data/query_data/KPR--EN 01-ROW-INDIE')
#     print "get_kw_duplicates some data"
#     dd_df = get_kw_duplicates(df)
#     print "saving result"
#     df.save('/D/Dropbox/dev/py/data/query_data/dedup_KPR--EN 01-ROW-INDIE')





__author__ = 'thorwhalen'

import pandas as pd
import ut.util.ulist as ulist
import ut.pstr.trans as pstr_trans
import ut.aw.manip as aw_manip

_Broad = 'BROAD'
_Phrase = 'PHRASE'
_Exact = 'EXACT'
_match_type_list =  [_Broad,_Phrase,_Exact]
_match_type_tag = {_Broad:'B', _Phrase:'P', _Exact:'X'}
_no_match_tag = '_'
_bpx_types = ['BPX','_PX','__X','BP_','B_X','_P_','B__']

def bid_order_stats(df):
    # count the occurances of the 7 types of BPXs
    stats = dict()
    num_of_rows = len(df)
    stats['bid_ordered'] = sum(df['bid_ordered'])
    stats['BP_bid_ordered'] = sum(df['BP_bid_ordered'])
    stats['PX_bid_ordered'] = sum(df['PX_bid_ordered'])
    stats['BX_bid_ordered'] = sum(df['BX_bid_ordered'])
    return pd.DataFrame(pd.Series(stats),columns=['count'])

def bpx_stats(df):
    # count the occurances of the 7 types of BPXs
    stats_df = pd.DataFrame(df[['bpx_tag']].groupby(['bpx_tag']).size(),columns=['count'])
    # add any BPXs that may be missing
    missing_bpx_types = list(set(_bpx_types)-set(stats_df.index))
    if len(missing_bpx_types) > 0:
        stats_df = stats_df.append(pd.DataFrame(pd.Series(data=0, index=missing_bpx_types),columns=['count']))
    stats_df.columns = ['count']
    # add a ratio column
    n_rows = len(df)
    stats_df['ratio'] = stats_df.apply(lambda x: x/float(n_rows))
    # reorder indices to the _bpx_types order
    stats_df = stats_df.reindex(_bpx_types)
    return stats_df

def add_bpx_col(df, groupby_keys=[]):
    groupby_keys = ulist.ascertain_list(groupby_keys) + ['kw_stripped_and_lowered']
    df['kw_stripped_and_lowered'] = pstr_trans.lower(aw_manip.strip_kw(df['keyword']))
    dg = df.groupby(groupby_keys,group_keys=False).apply(lambda x:_bpx_tag(x))
    del dg['kw_stripped_and_lowered']
    return dg

def _bpx_tag(grp):
    match_types = list(grp.match_type)
    bpx_tag = '' # initialize tag string
    bids_by_match = dict()
    for t in _match_type_list:
        if t in match_types:
            bpx_tag = bpx_tag + _match_type_tag[t]
            bids_by_match[t] = list(grp.max_cpc[grp.match_type==t])
        else:
            bpx_tag = bpx_tag + _no_match_tag
            bids_by_match[t] = []
    grp['bpx_tag'] = bpx_tag
    grp['BP_bid_ordered'] = True
    grp['PX_bid_ordered'] = True
    grp['BX_bid_ordered'] = True
    if bids_by_match[_Phrase]:
        grp['BP_bid_ordered'] = max(bids_by_match[_Broad]+bids_by_match[_Phrase]) == max(bids_by_match[_Phrase])
    if bids_by_match[_Exact]:
        grp['PX_bid_ordered'] = max(bids_by_match[_Phrase]+bids_by_match[_Exact]) == max(bids_by_match[_Exact])
        grp['BX_bid_ordered'] = max(bids_by_match[_Broad]+bids_by_match[_Exact]) == max(bids_by_match[_Exact])
    grp['bid_ordered'] = grp['BP_bid_ordered'] & grp['PX_bid_ordered'] & grp['BX_bid_ordered']
    return grp



def split_dolls(df):
    return df.groupby(['ad_group'],group_keys=False).apply(lambda x:_split_fun(x))

def _split_fun(grp):
    d = pd.concat([_rep_ad_group_name(grp[grp.match_type==match],match) for match in _match_type_list])
    d['target'] = d.match_type
    return d

def _rep_ad_group_name(d,match_type):
    d['ad_group'] =  [_ag_split_name(x,match_type) for x in d['ad_group']]
    return d

def _ag_split_name(ag_name,match_type):
    return ag_name + '|' + match_type









__author__ = 'thorwhalen'
"""
Includes functions to manipulate adwords datasets
"""

import pandas as pd
import numpy as np
import re
import ut as ms

from ut.daf.check import has_columns
#import ut.util
from ut.daf.manip import lower_series
from ut.daf.manip import select_col_equals
#from ut.util.ulist import ascertain_list
import ut.pstr.trans as pstr_trans

import ut.util.ulist as util_ulist
import ut.pcoll.order_conserving as colloc
import ut.daf.manip as daf_manip
import ut.aw.reporting

from ut.util.ulist import get_first_item_contained_in_intersection_of

tokenizer_re = re.compile('[&\w]+')
non_w_re = re.compile('\W+')


def process_aw_column_names(df):
    """
    Processes (in place) column names for aw name normalization.
    It replaces some synonyms (using ut.aw.reporting.x_to_lu_name()), lower cases them, and
    replaces all non \w characters by an underscore.
    """
    df.columns = ms.aw.reporting.x_to_lu_name(list(df.columns))
    df.columns = [x.lower() for x in df.columns]
    df.columns = [non_w_re.sub('_', x) for x in df.columns]
    df.columns = ms.aw.reporting.x_to_lu_name(list(df.columns))


def process_aw_column_values(df):
    """
    Returns the df with column values in their expected format.
    (See below his py file for column_name -> processing_function map)
    """
    for col in df.columns:
        try:
            df[col] = column_names_to_preproc_function[col](df[col])
        except KeyError:
            continue
    return df


def figure_out_match_type_from_keywords_and_simplify_keywords(df):
    df = df.copy()
    if 'match_type' in df.columns:
        raise RuntimeWarning("there's a column named match_type in your df, so to be safe, I'm not going to infer it")
        return df
    elif 'keyword' not in df.columns:
        raise RuntimeError("You don't have a keyword column. I need that!")
    else:
        kw_list = np.array(df['keyword'])
        df['match_type'] = 'broad'
        for i in range(len(kw_list)):
            kw = kw_list[i]
            if kw[0] == '[':
                df['match_type'].iloc[i] = 'exact'
                df['keyword'].iloc[i] = kw[1:-1]  # remove first and last characters (should be [ and ])
            elif kw[0] == '"':
                df['match_type'].iloc[i] = 'phrase'
                df['keyword'].iloc[i] = kw[1:-1]  # remove first and last characters (should be " and ")
        return df


def mk_terms_df(df, text_cols, id_cols=None, tokenizer_re=tokenizer_re):
    text_cols = util_ulist.ascertain_list(text_cols)
    df = df.copy()
    for col in text_cols:
        df[col] = kw_str(df[col])
    return semantics_term_stats_maker_mk_terms_df(df, text_cols, id_cols, tokenizer_re)


def semantics_term_stats_maker_mk_terms_df(df, text_cols, id_cols=None, tokenizer_re=tokenizer_re):
    text_cols = util_ulist.ascertain_list(text_cols)
    if id_cols is None:
        id_cols = colloc.setdiff(df.columns, text_cols)
    else:
        id_cols = util_ulist.ascertain_list(id_cols)
        id_cols_missing = colloc.setdiff(id_cols, df.columns)
        if id_cols_missing: # if any columns are missing, try to get them from named index
            df = df.reset_index(id_cols_missing)
    dd = pd.DataFrame()
    for c in text_cols:
        d = df[id_cols]
        d['term'] = [re.findall(tokenizer_re, x) for x in df[c]]
        d = daf_manip.rollout_cols(d, cols_to_rollout='term')
        dd = pd.concat([dd, d])
    return dd


def kw_str(keyword):
    """
        produces a kw_str version of the input keyword (or list of keywords), i.e. lower ascii and strip_kw are applied
    """
    #return strip_kw(pstr_trans.lower(pstr_trans.toascii(pstr_trans.to_unicode_or_bust(keyword))))
    if isinstance(keyword, str):
        return str(
            strip_kw(
                pstr_trans.lower(
                    pstr_trans.toascii(
                        pstr_trans.to_unicode_or_bust(keyword)))))
    else:
        return [str(
            strip_kw(
                pstr_trans.lower(
                    pstr_trans.toascii(
                        pstr_trans.to_unicode_or_bust(x))))) for x in keyword]


def strip_kw(keyword):
    """
        replaces keywords (a single string or list thereof) by a version of the string that contains only characters that
        are considered by google adwords (namely, letters, numbers, and & (and underscore too, which I left hoping they'd
        never show up (they shouldn't since google doesn't allow them, or if they do, that we can just leave these)
    """
    exp = re.compile('[^\w&]', re.UNICODE)
    if isinstance(keyword, str):
        return ' '.join(re.sub(exp, ' ', keyword).split())
    else: # assume it's an iterable collection of keywords
        return [' '.join(re.sub(exp, ' ', kw).split()) for kw in keyword]

def order_words(keyword):
    """
        orders the words of a keyword string (or iterable thereof).
        NOTE: words are here defined as any string of non-space characters (you may want to consider stripping the keyword first)
    """
    #exp = re.compile('[^\w&]', re.UNICODE)
    try: # assume keyword is a string
        return ' '.join(np.sort(keyword.split(' ')))
    except: # assume it's an iterable collection of keywords
        return [' '.join(np.sort(kw.split(' '))) for kw in keyword]


def add_col(df, colname=None, overwrite=True, **kwargs):
    """
    Adds one or several requested columns (colname) to df, usually computed based on other columns of the df.
    Details of what colname does what inside the code!
    The overwrite flag (defaulted to True) specified whether
    """
    if colname is None:
        print("colname choices: ")
        print("%s" % str(['pos_impressions', 'pos', 'day_of_week_num', 'day_of_week', 'week_of_year',
                          'cvr', 'ctr', 'cpc', 'spc', 'kw_lower', 'kw_lower_ascii', 'kw_lower_ascii_ordered',
                          'destination']))
        return None
    df_columns = df.columns
    if isinstance(colname, str):
        if overwrite is False and has_columns(df, colname):
            return df # just return the df as is
        else:
            if colname == 'pos_impressions':
                df['pos_impressions'] = df['avg_position'] * df['impressions']
            elif colname == 'pos':
                df['pos'] = df['pos_impressions'] / df['impressions']
            elif colname == 'day_of_week_num':
                if 'day' in df.columns:
                    df['day_of_week_num'] = df['day'].apply(pd.datetime.weekday)
                elif 'date' in df.columns:
                    df['day_of_week_num'] = df['date'].apply(pd.datetime.weekday)
                else:
                    days = ['Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday', 'Sunday']
                    key_col = 'day_of_week'
                    day_2_num = pd.DataFrame({'day_of_week': days, 'day_of_week_num': np.arange(len(days))})
                    if key_col in df.index.names:
                        index_names = df.index.names
                        df = df.reset_index(drop=False)
                        df = df.merge(day_2_num)
                        if kwargs.get('rm_key_cols', False):
                            df.drop(key_col, axis=1, inplace=True)
                            index_names = list(set(index_names).difference([key_col]))
                        df = df.set_index(index_names)
                    else:
                        df = df.merge(day_2_num)
                        if kwargs.get('rm_key_cols', False):
                            df.drop(key_col, axis=1, inplace=True)
            elif colname == 'day_of_week':
                days = ['Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday', 'Sunday']
                key_col = 'day_of_week_num'
                day_2_num = pd.DataFrame({'day_of_week': days, 'day_of_week_num': np.arange(len(days))})
                if key_col in df.index.names:
                    index_names = df.index.names
                    df = df.reset_index(drop=False)
                    df = df.merge(day_2_num)
                    if kwargs.get('rm_key_cols', False):
                        df.drop(key_col, axis=1, inplace=True)
                        index_names = list(set(index_names).difference([key_col]))
                    df = df.set_index(index_names)
                else:
                    df = df.merge(day_2_num)
                    if kwargs.get('rm_key_cols', False):
                        df.drop(key_col, axis=1, inplace=True)
            elif colname == 'week_of_year':
                date_col = kwargs.get('date_col', None)
                if date_col is None:
                    date_col = get_first_item_contained_in_intersection_of(['day', 'date'], df.columns, None)
                if date_col is None:
                    raise KeyError("Couldn't find a date_col to work with: Tell me what it is")
                if isinstance(date_col, str):
                    date_col = df[date_col]
                try:
                    df['week_of_year'] = [t.isocalendar()[1] for t in date_col]
                except AttributeError:
                    df['week_of_year'] = [t.weekofyear for t in date_col]

            elif colname == 'cvr':
                df['cvr'] = df['conversions'] / df['clicks']
            elif colname == 'ctr':
                df['ctr'] = df['clicks'] / df['impressions']
            elif colname == 'cpc':
                df['cpc'] = df['cost'] / df['clicks']
            elif colname == 'spc':
                mean_cvr = sum(df['conversions']) / sum(df['clicks'])
                prior_clicks = kwargs.get('prior_clicks', 300)
                df['spc'] = (df['conversions'] + mean_cvr * prior_clicks) / (df['clicks'] + prior_clicks)
            elif colname == 'kw_lower':
                assert_dependencies(df, 'keyword', "to get {}".format(colname))
                df[colname] = lower_series(df['keyword'])
            elif colname == 'kw_lower_ascii':
                assert_dependencies(df, 'keyword', "to get {}".format(colname))
                df[colname] = pstr_trans.toascii(lower_series(df['keyword']))
            elif colname == 'kw_lower_ascii_ordered':
                assert_dependencies(df, 'keyword', "to get {}".format(colname))
                df[colname] = [' '.join(np.sort(x.split(' '))) for x in pstr_trans.toascii(lower_series(df['keyword']))]
            elif colname == 'destination':
                if 'ad_group' in df_columns:
                    #ag_triad = map(lambda x: x.split('|'), pstr_trans.lower(pstr_trans.toascii(list(df['ad_group']))))
                    ag_triad = [x.split('|') for x in df['ad_group']]
                    ag_triad_0 = kw_str([x[0] for x in ag_triad])
                    ag_triad_2 = kw_str([x[2] for x in ag_triad])
                    df[colname] = list(map(lambda x2, x0: '|'.join([x2, x0]), ag_triad_2, ag_triad_0))
                elif 'campaign' in df_columns:
                    df[colname] = kw_str(df['campaign'])
                else:
                    raise ValueError('You need ad_group or campaign to get a destination')
            else:
                raise RuntimeError("unknown colname requested")
            # remove columns?
            if 'remove_cols' in list(kwargs.keys()):
                df.drop(set(kwargs.get('remove_cols', None)).union(df.columns), axis=1, inplace=True)
    else:
        try:
            for c in colname:
                df = add_col(df, c, overwrite=overwrite)
        except TypeError:
            raise RuntimeError("colname must be a string or a list of string")
    return df


def get_kw_active(df):
    """
    returns the df with only enabled keyword status
    """
    return select_col_equals(df, 'keyword_state','enabled')


def assert_dependencies(df, cols, prefix_message=""):
    assert has_columns(df, cols), "need (all) columns {}: {}".format(util_ulist.to_str(cols), prefix_message)


# def adjust_bid(val, min_bid=0.01, max_bid=20.0, f=lambda x: x):
#     """
#     Adjusts the bids based on passed in function and ensures no bid is too low or too high.
#     Useful for operations on dataframe columns, such as when adjusting bids.
#
#     example: kw_df.max_cpc = kw_df.max_cpc.apply(adjust_bid, f=lambda x: x*2)
#     """
#     new_val = f(val)
#     new_val = min_bid if new_val < min_bid else new_val
#     new_val = max_bid if new_val > max_bid else new_val
#     return new_val

def _percentage_str_to_ratio_float(col):
    return col.apply(lambda x: x[:-1]).apply(float) / 100.

def _date_str_to_datetime(col):
    return col.apply(pd.to_datetime)

column_names_to_preproc_function = {
    'ctr': _percentage_str_to_ratio_float,
    'conv_rate': _percentage_str_to_ratio_float,
    'day': _date_str_to_datetime,
    'date': _date_str_to_datetime,
    'hour_of_day': lambda x: x.apply(int)
}

columns_that_should_be_floats = [
    'avg_position', 'impressions', 'clicks', 'conversions', 'cost_conv_', 'view_through_conv',
    'cost', 'cpc'
]

for col in columns_that_should_be_floats:
    column_names_to_preproc_function.update({col: lambda x: x.apply(float)})



__author__ = 'thorwhalen'

import datetime
from . import reporting as rp
import pandas

#settings
save_folder = '/D/Dropbox/dev/py/data/query_data/'
account_list = rp.get_account_id('dict')
account_list  = list(account_list.keys())
numOfDays = 60
report_query_str = rp.mk_report_query_str(
    varList='q_iipic',
    start_date= numOfDays)
def save_file(account):
    return "{}{}-{}-{}days.p".format(
        save_folder,
        account,
        datetime.date.today().strftime("%Y%m%d"),
        numOfDays)


def getting_a_bunch_of_queries():
        # running that shit
    i = 0
    account_list = ['test']
    for account in account_list:
        i = i + 1
        print("({}/{}) downloading {}".format(i,len(account_list),account))
        report_downloader = rp.get_report_downloader(account)
        df = rp.download_report(
            report_downloader=report_downloader,
            report_query_str=report_query_str,
            download_format='df')
    saveFile = save_file(account)
    print("   Saving to {}".format(saveFile))
    df.save(saveFile)


# run this shit
getting_a_bunch_of_queries()


__author__ = 'thorwhalen'

import pfile.accessor as pfile_accessor
from analyzer.pstore import MyStore
from analyzer.pstore import StoreAccessor

local_facc = pfile_accessor.for_local('hdf5/')


def for_local(
        store_path_dict=None
):
    if store_path_dict is None:
        store_path_dict = {
            'ad_elements': local_facc('aw/ad_elements.h5'),
            'kw_elements': local_facc('aw/kw_elements.h5'),
        }
        add_from_dict = {
            'ad_group': {'join_store': 'ad_elements', 'join_cols': 'ad_group_id'}
        }
    return StoreAccessor(store_path_dict=store_path_dict, add_from_dict=add_from_dict)

def from_ad_and_kw_elements(
        store_path_dict=None
):
    if store_path_dict is None:
        store_path_dict = {
            'ad_elements': local_facc('aw/ad_elements.h5'),
            'kw_elements': local_facc('aw/kw_elements.h5'),
        }
        add_from_dict = {
            'ad_group': {'join_store': 'ad_elements', 'join_cols': 'ad_group_id'}
        }
    return StoreAccessor(store_path_dict=store_path_dict, add_from_dict=add_from_dict)


def from_stpr(
        store_path_dict=None
):
    if store_path_dict is None:
        store_path_dict = {
            'stpr': local_facc('stpr/stpr_20130107_20130507.h5')
        }
        add_from_dict = {
            'ad_group': {'join_store': 'stpr', 'join_key': 'stpr_20130107_20130507', 'join_cols': 'ad_group_id'},
            'keyword': {'join_store': 'stpr', 'join_key': 'stpr_20130107_20130507', 'join_cols': ['ad_group_id', 'search_term']}
        }
    return StoreAccessor(store_path_dict=store_path_dict, add_from_dict=add_from_dict)


__author__ = 'thorwhalen'


import ut.aw
from numpy.lib import arraysetops
import pandas as pd

import ut.daf
import ut.util


def get_unique(d,cols):
    d = d.reindex(index=list(range(len(d))))
    grouped = d.groupby(cols)
    index = [gp_keys[0] for gp_keys in list(grouped.groups.values())]
    return d.reindex(index)


def get_non_low_str_of_col(d,col):
    return d[d[col] != ut.aw.manip.lower_series(d[col])]


def print_unique_counts(d):
    column_list = d.columns.tolist()
    print("number of rows: {}".format(len(d[column_list[0]])))
    print("")
    for c in column_list:
        print("number of unique {}: {}".format(c,len(arraysetops.unique(d[c]))))


def get_kw_duplicates(df):
    df =ut.aw.manip.add_col(df,'kw_lower',overwrite=False)
    df = get_duplicates(df,'kw_lower')


def get_duplicates(df,cols):
    df = df.reindex(index=list(range(len(df))))
    grouped = df.groupby(cols)
    unique_index = [gp_keys[0] for gp_keys in list(grouped.groups.values())]
    non_unique_index = list(set(df.index)-set(unique_index))
    duplicates_df = ut.aw.diagnosis.get_unique(df.irow(non_unique_index),cols)
    duplicates_df = duplicates_df[cols]
    return df.merge(pd.DataFrame(duplicates_df))


def add_col(df,colname,overwrite=True):
    """
    Adds one or several requested columns (colname) to df, usually computed based on other columns of the df.
    Details of what colname does what inside the code!
    The overwrite flag (defaulted to True) specified whether
    """
    if isinstance(colname,str):
        if overwrite==False and ut.daf.check.has_columns(df,colname):
            return df # just return the df as is
        else:
            if colname=='lower_kw':
                assert_dependencies(df,'Keyword',"to get {}".format(colname))
                df['lower_kw'] = lower_series(df['Keyword'])
            else:
                raise RuntimeError("unknown colname requested")
    elif isinstance(colname,list):
        for c in colname:
            df = add_col(df,c,overwrite=overwrite)
    else:
        raise RuntimeError("colname must be a string or a list of string")
    return df

def index_with_range(df):
    return df.reindex(index=list(range(len(df))))

def lower_series(sr):
    return sr.apply(lambda x: x.lower())

def lower_column(df,col):
    df[col] = lower_series(df[col])
    return df

def get_kw_active(df):
    """
    returns the df with only enabled keyword status
    """
    return select_col_equals(df,'Keyword state','enabled')

def select_col_equals(df,col,value):
    """
    returns the df with only enabled keyword status
    """
    assert_dependencies(df,col)
    return index_with_range(df[df[col]==value])

def rm_nan_rows(df):
    """
    removes all rows containing any nans
    """
    return index_with_range(df.dropna())


def assert_dependencies(df,cols,prefix_message=""):
    assert ut.daf.check.has_columns(df,cols),"need (all) columns {}: {}".format(ut.util.ulist.to_str(cols),prefix_message)


__author__ = 'thorwhalen'

import numpy as np
from ut.datapath import datapath
import ut.daf.diagnosis as daf_diagnosis
import ut.parse.google as google
import ut.parse.util as parse_util
import pandas as pd
import ut.daf.ch as daf_ch

import os
import re
from ut.pstr.trans import toascii
from . import reporting
from datetime import datetime
from ut.util.ulist import ascertain_list
from ut.util.ulist import all_true
from ut.pstr.trans import to_unicode_or_bust
from serialize.data_accessor import DataAccessor

split_exp = re.compile("[^&\w]*")

travel_domain_list = ['expedia', 'tripadvisor', 'trivago', 'marriott', 'booking', 'hotels', 'lastminute', 'accorhotels',
                      'kayak', 'venere', 'hilton', 'hotelscombined', 'agoda', 'choicehotels', 'travelocity',
                      'travelsupermarket', 'bestwestern', 'laterooms', 'radissonblu', 'hotwire', 'lonelyplanet',
                      'orbitz', 'starwoodhotels', 'frommers', 'hotel', 'hotelclub', 'hrs', 'novotel', 'wego', 'wotif',
                      'hoteltravel', 'hyatt', 'ibis', 'ihg', 'mercure', 'priceline', 'qualityinn',
                      'beoo', 'easytobook', 'ebookers', 'hostelbookers', 'lq', 'melia', 'millenniumhotels', 'mrandmrssmith',
                      'nh-hotels', 'ratestogo', 'sofitel', 'tablethotels', 'travelandleisure']

html_data = DataAccessor('html/google_results_tests')

def get_search_term_html(search_term):
    file_name = html_data+'.html'
    try:
        return html_data.loads(file_name)
    except:
        IOError("didn't find %s" % html_data.filepath(file_name))

def google_light_parse(gresult):
    gresult = parse_util.x_to_soup(gresult)
    parse_dict = dict()
    resultStats = input.find(name='div',attrs={'id':'resultStats'})
    if resultStats:
        parse_dict['_resultStats'] = google.parse_number_of_results(resultStats)




def save_search_term_that_does_not_have_num_of_results(search_term):
    print("no num_of_results in: %s" % search_term)

def add_travel_score(query_report_df, html_folder=datapath(),ext='.html'):
    pass


# TODO: Continue coding add_travel_score()
def mk_search_term_domains_df(query_report_df, html_folder=datapath(),ext='.html'):
    search_terms = np.unique(list(query_report_df['search_term']))
    domain_lists = []
    search_term_list = []
    for st in search_terms:
        filename = os.path.join(html_folder,st+ext)
        print(filename)
        if os.path.exists(filename):
            search_term_list.append(st)
            domain_lists.append(get_domain_list_from_google_results(filename))
    return pd.DataFrame({'search_term':search_term_list, 'domain_list':domain_lists})

def add_search_term_ndups(df, count_var='ndups'):
    d = df[['search_term']].groupby('search_term').count()
    d.columns = [count_var]
    return df.merge(d, left_on='search_term', right_index=True)

def add_target_scores(query_report_df):
    vars_to_keep = ['search_term','impressions','destination','ad_group','destination_imps_freq_fanout_ratio','ad_group_imps_freq_fanout_ratio']
    query_report_df = add_query_fanout_scores(query_report_df)
    query_report_df = query_report_df[vars_to_keep]
    query_report_df = daf_ch.ch_col_names(query_report_df,
                                   ['ad_group_score','destination_score'],
                                   ['ad_group_imps_freq_fanout_ratio','destination_imps_freq_fanout_ratio'])
    query_report_df = query_report_df.sort(columns=['search_term','destination_score','ad_group_score'])
    return query_report_df

def add_query_fanout_scores(query_report_df):
    if 'destination' not in query_report_df.columns:
        query_report_df = add_destination(query_report_df)
    ad_group_fanout = mk_query_fanout_scores(query_report_df,target='ad_group',statVars='impressions',keep_statVars=False)
    ad_group_fanout = daf_ch.ch_col_names(ad_group_fanout,
                                   ['ad_group_imps_freq_fanout_ratio','ad_group_count_fanout_ratio'],
                                   ['impressions_freq_fanout_ratio','impressions_count_fanout_ratio'])
    destination_fanout = mk_query_fanout_scores(query_report_df,target='destination',statVars='impressions',keep_statVars=False)
    destination_fanout = daf_ch.ch_col_names(destination_fanout,
                                   ['destination_imps_freq_fanout_ratio','destination_count_fanout_ratio'],
                                   ['impressions_freq_fanout_ratio','impressions_count_fanout_ratio'])
    query_report_df = query_report_df.merge(ad_group_fanout,on=['search_term','ad_group'])
    query_report_df = query_report_df.merge(destination_fanout,on=['search_term','destination'])
    return query_report_df


def mk_query_fanout_scores(query_report_df, target='ad_group', statVars='impressions',keep_statVars=False):
    # target = ascertain_list(target)
    # if ('destination' in target) and ('destination' not in query_report_df.columns):
    #     query_report_df['destination'] = query_report_df['ad_group'].apply(lambda x : re.match('[^|]*',x).group(0))
    # if not all_true([x in query_report_df.columns for x in target]):
    #     raise ValueError("the dataframe doesn't have the column %s and I don't know how to make it" % target)
    if target not in query_report_df.columns:
        if target=='destination':
            query_report_df = add_destination(query_report_df)
        else:
            raise ValueError("the dataframe doesn't have the column %s and I don't know how to make it" % target)
    return daf_diagnosis.mk_fanout_score_df(query_report_df,fromVars=['search_term'],toVars=target,statVars=statVars,keep_statVars=keep_statVars)

def add_destination(query_report_df):
    assert 'ad_group' in query_report_df.columns, "you must have the variable ad_group to infer destination"
    # get the destination as the substring before the first |
    query_report_df['destination'] = query_report_df['destination'] = \
        query_report_df['ad_group'].apply(lambda x : re.match('[^|]*',x).group(0))
    return query_report_df

def get_domain_list_from_google_results(gresults):
    domain_list = []
    if not isinstance(gresults,dict): # assume it's a soup, html, or filename thereof
        gresults = google.parse_tag_dict(google.mk_gresult_tag_dict(gresults))
    # if not, assume the input is a info_dict
    if 'organic_results_list' in gresults:
        domain_list = domain_list + [x['domain'] for x in gresults['organic_results_list'] if 'domain' in x]
    if 'top_ads_list' in gresults:
        domain_list = domain_list + [x['disp_url_domain'] for x in gresults['top_ads_list'] if 'disp_url_domain' in x]
    if 'organic_results_list' in gresults:
        domain_list = domain_list + [x['disp_url_domain'] for x in gresults['organic_results_list'] if 'disp_url_domain' in x]
    return domain_list


def similarity_with_travel_domains(domain_list):
    if isinstance(domain_list, pd.DataFrame):
        domain_list = list(domain_list.index)
    return len(set(domain_list).intersection(set(travel_domain_list)))\
               / float(len(domain_list))



###############################################################################################
## SEMANTIC STUFF


def mk_term_count_from_google_results(gresults):
    """
    takes a google result (in the form of html, filename thereof, soup, or info_dict, and
    returns a Series whose indices are terms and values are term counts
    (the number of times the term appeared in the google result)
    """
    # get preprocessed text from gresults
    gresults = process_text_for_word_count(mk_text_from_google_results(gresults))
    # tokenize this text
    toks = tokenize_text(gresults)
    # make a dataframe of term counts TODO: Explore faster ways to do this
    df = pd.DataFrame(toks,columns=['token'])
    df = df.groupby('token').count()
    df.columns = ['count']
    df = df.sort(columns=['count'],ascending=False) # TODO: Take out sorting at some point since it's unecessary (just for diagnosis purposes)
    return df

def tokenize_text(gresult_text):
    return re.split(split_exp,gresult_text)

def process_text_for_word_count(text):
    """
    Preprocesses the text before it will be fed to the tokenizer.
    Here, we should put things like lower-casing the text, casting letters to "simple" ("ascii", "non-accentuated")
    letters, replacing some common strings (such as "bed and breakfast", "New York" by singular token representatives
    such as "b&b", "new_york"), and what ever needs to be done before tokens are retrieved from text.
    """
    return toascii(to_unicode_or_bust(text)).lower()


def mk_text_from_google_results(gresults):
    if not isinstance(gresults,dict): # if not a dict assume it's a soup, html, or filename thereof
        gresults = google.parse_tag_dict(google.mk_gresult_tag_dict(gresults))
    if 'organic_results_list' in gresults:
        title_text_concatinated = ' '.join([x['title_text'] for x in gresults['organic_results_list'] if 'title_text' in x])
        snippet_text_concatinated = ' '.join([x['st_text'] for x in gresults['organic_results_list'] if 'st_text' in x])
        text_concatinated = title_text_concatinated + ' ' + snippet_text_concatinated
    else:
        search_for_tag = ['_ires','_search','_res','_center_col']
        for t in search_for_tag:
            if t in gresults:
                text_concatinated = soup_to_text(gresults[t])
                break
        if not text_concatinated: # if you still don't have anything
            text_concatinated = soup_to_text(gresults) #... just get the text from the whole soup
    return text_concatinated

def soup_to_text(element):
    return list(filter(visible, element.findAll(text=True)))

def visible(element):
    if element.parent.name in ['style', 'script', '[document]', 'head', 'title']:
        return False
    elif re.match('<!--.*-->', str(element)):
        return False
    return True



###############################################################################################
## MAKING DATA

def mk_df_of_travel_domains():
    # set up resources
    html_folder = '/D/Dropbox/dev/py/data/html/google_results_tests/'
    file_list = ['hotel - 100 Google Search Results.html',
             'find hotel deals - 100 Google Search Results.html',
             'hotel travel sites - 100 Google Search Results.html',
             'find hotels - 100 Google Search Results.html',
             'hotel paris - 100 Google Search Results.html',
             'hotel rome - 100 Google Search Results.html',
             'hotel london - 100 Google Search Results.html',
             'hotel nyc - 100 Google Search Results.html',
             'hotels in france - 100 Google Search Results.html',
             'hotels in italy - 100 Google Search Results.html'
            ]
    filepath_list = [os.path.join(html_folder,f) for f in file_list]
    # parse all this
    r = [google.mk_gresult_tag_dict(f) for f in filepath_list]
    r = [google.parse_tag_dict(f) for f in r]
    # make domain lists
    org_domain_list = []
    ads_domain_list = []
    tads_domain_list = []
    for rr in r:
        rrr = rr['organic_results_list']
        org_domain_list = org_domain_list + [x['domain'] for x in rrr if 'domain' in x]
        rrr = rr['rhs_ads_list']
        ads_domain_list = ads_domain_list + [x['disp_url_domain'] for x in rrr if 'disp_url_domain' in x]
        rrr = rr['top_ads_list']
        ads_domain_list = ads_domain_list + [x['disp_url_domain'] for x in rrr if 'disp_url_domain' in x]
    domain_list = org_domain_list + ads_domain_list
    print("number of org_domain_list entries = %d" % len(org_domain_list))
    print("number of ads_domain_list entries = %d" % len(ads_domain_list))
    print("number of (all) domain_list entries = %d" % len(domain_list))
    # make a dataframe counting the number of times we encouter each domain
    df = pd.DataFrame(domain_list,columns=['domain'])
    dg = df.groupby('domain').count() #agg([('domain_count','len')])
    dg = daf_ch.ch_col_names(dg,'count','domain')
    thresh = 4
    print("length before removing count<%d entries = %d" % (thresh,len(dg)))
    dg = dg[dg['count']>=thresh]
    print("length before removing count<%d entries = %d" % (thresh,len(dg)))
    dg['frequency'] = dg['count']/float(max(dg['count']))
    dg = dg.sort(columns=['count'],ascending=False)
    dg.head(30)
    # return this!
    return dg





__author__ = 'thorwhalen'





lu_name_x_dict = {
    'a_ce_split': ['aCESplit', 'ACE split'],
    'account': ['account', 'Account'],
    'account_id': ['accountID', 'Account ID'],
    'ad': ['ad', 'Ad'],
    'ad_approval_status': ['adApprovalStatus', 'Ad Approval Status'],
    'ad_extension_id': ['adExtensionID', 'Ad Extension ID'],
    'ad_extension_type': ['adExtensionType', 'Ad Extension Type'],
    'ad_group': ['adGroup', 'Ad group'],
    'ad_group_id': ['adGroupID', 'Ad group ID'],
    'ad_group_state': ['adGroupState', 'Ad group state'],
    'ad_id': ['adID', 'Ad ID'],
    'ad_state': ['adState', 'Ad state'],
    'ad_type': ['adType', 'Ad type'],
    'added': ['added', 'Added'],
    'approval_status': ['approvalStatus', 'Approval Status'],
    'attribute_values': ['attributeValues', 'Attribute Values'],
    'audience': ['audience', 'Audience'],
    'audience_state': ['audienceState', 'Audience state'],
    'avg_cpc': ['avgCPC', 'Avg. CPC'],
    'avg_cpm': ['avgCPM', 'Avg. CPM'],
    'avg_cpp': ['avgCPP', 'Avg. CPP'],
    'avg_position': ['avgPosition', 'Avg. position'],
    'bidding_strategy': ['biddingStrategy', 'Bidding strategy'],
    'budget': ['budget', 'Budget'],
    'budget_explicitly_shared': ['budgetExplicitlyShared',
                                  'Budget explicitly shared'],
    'budget_id': ['budgetID', 'Budget ID'],
    'budget_name': ['budgetName', 'Budget Name'],
    'budget_period': ['budgetPeriod', 'Budget period'],
    'budget_state': ['budgetState', 'Budget state'],
    'budget_usage': ['budgetUsage', 'Budget usage'],
    'business_phone_number': ['businessPhoneNumber', 'Business phone number'],
    'cpc_ace_indicator': ['cPCACEIndicator', 'CPC ACE indicator'],
    'cpm_ace_indicator': ['cPMACEIndicator', 'CPM ACE indicator'],
    'ctr_ace_indicator': ['cTRACEIndicator', 'CTR ACE indicator'],
    'call_fee': ['callFee', 'Call fee'],
    'caller_area_code': ['callerAreaCode', 'Caller area code'],
    'caller_country_code': ['callerCountryCode', 'Caller country code'],
    'campaign': ['campaign', 'Campaign'],
    'campaign_id': ['campaignID', 'Campaign ID'],
    'campaign_state': ['campaignState', 'Campaign state'],
    'campaigns': ['campaigns', '# Campaigns'],
    'categories': ['categories', 'Categories'],
    'city': ['city', 'City'],
    'click_id': ['clickId', 'Click Id'],
    'click_type': ['clickType', 'Click type'],
    'clicks': ['clicks', 'Clicks'],
    'clicks_ace_indicator': ['clicksACEIndicator', 'Clicks ACE indicator'],
    'client_name': ['clientName', 'Client name'],
    'company_name': ['companyName', 'Company name'],
    'content_impr_share': ['contentImprShare', 'Content Impr. share'],
    'content_lost_is_budget': ['contentLostISBudget',
                                'Content Lost IS (budget)'],
    'content_lost_is_rank': ['contentLostISRank', 'Content Lost IS (rank)'],
    'conv': ['conv', 'Conv.'],
    'conv1_per_click': ['conv1PerClick', 'Conv. (1-per-click)'],
    'conv1_per_click_ace_indicator': ['conv1PerClickACEIndicator',
                                       'Conv. (1-per-click) ACE indicator'],
    'conv_many_per_click': ['convManyPerClick', 'Conv. (many-per-click)'],
    'conv_many_per_click_ace_indicator': ['convManyPerClickACEIndicator',
                                           'Conv. (many-per-click) ACE indicator'],
    'conv_rate': ['convRate', 'Conv. rate'],
    'conv_rate1_per_click': ['convRate1PerClick', 'Conv. rate (1-per-click)'],
    'conv_rate1_per_click_ace_indicator': ['convRate1PerClickACEIndicator',
                                            'Conv. rate (1-per-click) ACE indicator'],
    'conv_rate_many_per_click': ['convRateManyPerClick',
                                  'Conv. rate (many-per-click)'],
    'conv_rate_many_per_click_ace_indicator': ['convRateManyPerClickACEIndicator',
                                                'Conv. rate (many-per-click) ACE indicator'],
    'conversion_action_name': ['conversionActionName',
                                'Conversion action name'],
    'conversion_optimizer_bid_type': ['conversionOptimizerBidType',
                                       'Conversion optimizer bid type'],
    'conversion_tracker_id': ['conversionTrackerId', 'Conversion Tracker Id'],
    'conversion_tracking_purpose': ['conversionTrackingPurpose',
                                     'Conversion tracking purpose'],
    'cost': ['cost', 'Cost'],
    'cost_ace_indicator': ['costACEIndicator', 'Cost ACE indicator'],
    'cost_conv1_per_click': ['costConv1PerClick',
                              'Cost / conv. (1-per-click)'],
    'cost_conv1_per_click_ace_indicator': ['costConv1PerClickACEIndicator',
                                            'Cost/conv. (1-per-click) ACE indicator'],
    'cost_conv_many_per_click': ['costConvManyPerClick',
                                  'Cost / conv. (many-per-click)'],
    'cost_conv_many_per_click_ace_indicator': ['costConvManyPerClickACEIndicator',
                                                'Cost/conv. (many-per-click) ACE indicator'],
    'country_territory': ['countryTerritory', 'Country/Territory'],
    'criteria_display_name': ['criteriaDisplayName', 'Criteria Display Name'],
    'criteria_type': ['criteriaType', 'Criteria Type'],
    'criterion_id': ['criterionID', 'Criterion ID'],
    'ctr': ['ctr', 'CTR'],
    'currency': ['currency', 'Currency'],
    'customer_id': ['customerID', 'Customer ID'],
    'day': ['day', 'Day'],
    'day_of_week': ['dayOfWeek', 'Day of week'],
    'default_max_cpc': ['defaultMaxCPC', 'Default max. CPC'],
    'delivery_method': ['deliveryMethod', 'Delivery method'],
    'description_line1': ['descriptionLine1', 'Description line 1'],
    'description_line2': ['descriptionLine2', 'Description line 2'],
    'destination_url': ['destinationURL', 'Destination URL'],
    'device': ['device', 'Device'],
    'device_preference': ['devicePreference', 'Device preference'],
    'display_network_max_cpc': ['displayNetworkMaxCPC',
                                 'Display Network max. CPC'],
    'display_url': ['displayURL', 'Display URL'],
    'domain': ['domain', 'Domain'],
    'duration_seconds': ['durationSeconds', 'Duration (seconds)'],
    'dynamic_ad_target': ['dynamicAdTarget', 'Dynamic ad target'],
    'dynamically_generated_headline': ['dynamicallyGeneratedHeadline',
                                        'Dynamically generated Headline'],
    'end_time': ['endTime', 'End time'],
    'enhanced': ['enhanced', 'Enhanced'],
    'enhanced_cpc_enabled': ['enhancedCPCEnabled', 'Enhanced CPC enabled'],
    'excluded': ['excluded', 'Excluded'],
    'exclusion': ['exclusion', 'Exclusion'],
    'explicitly_shared': ['explicitlyShared', 'Explicitly shared'],
    'feed_id': ['feedID', 'Feed ID'],
    'feed_item_id': ['feedItemID', 'Feed item ID'],
    'feed_item_status': ['feedItemStatus', 'Feed item status'],
    'feed_placeholder_type': ['feedPlaceholderType', 'Feed placeholder type'],
    'first_level_sub_categories': ['firstLevelSubCategories',
                                    'First level sub-categories'],
    'first_page_cpc': ['firstPageCPC', 'First page CPC'],
    'free_click_rate': ['freeClickRate', 'Free click rate'],
    'free_click_type': ['freeClickType', 'Free click type'],
    'free_clicks': ['freeClicks', 'Free clicks'],
    'frequency': ['frequency', 'Frequency'],
    'highest_position': ['highestPosition', 'Highest position'],
    'hour_of_day': ['hourOfDay', 'Hour of day'],
    'image_ad_name': ['imageAdName', 'Image ad name'],
    'image_hosting_key': ['imageHostingKey', 'Image hosting key'],
    'impressions': ['impressions', 'Impressions'],
    'impressions_ace_indicator': ['impressionsACEIndicator',
                                   'Impressions ACE indicator'],
    'invalid_click_rate': ['invalidClickRate', 'Invalid click rate'],
    'invalid_clicks': ['invalidClicks', 'Invalid clicks'],
    'is_negative': ['isNegative', 'Is negative'],
    'is_targetable': ['isTargetable', 'Is Targetable'],
    'keyword': ['keyword', 'Keyword'],
    'keyword_id': ['keywordID', 'Keyword ID'],
    'keyword_max_cpc': ['keywordMaxCPC', 'Keyword max CPC'],
    'keyword_placement': ['keywordPlacement', 'Keyword / Placement'],
    'keyword_placement_destination_url': ['keywordPlacementDestinationURL',
                                           'Keyword/Placement destination URL'],
    'keyword_placement_state': ['keywordPlacementState',
                                 'Keyword/Placement state'],
    'keyword_state': ['keywordState', 'Keyword state'],
    'keyword_text': ['keywordText', 'Keyword text'],
    'landing_page_title': ['landingPageTitle', 'Landing Page Title'],
    'location': ['location', 'Location'],
    'location_extension_source': ['locationExtensionSource',
                                   'Location Extension Source'],
    'location_type': ['locationType', 'Location type'],
    'login_email': ['loginEmail', 'Login email'],
    'lowest_position': ['lowestPosition', 'Lowest position'],
    'match_type': ['matchType', 'Match type'],
    'max_cpa': ['maxCPA', 'Max. CPA%'],
    'max_cpa1_per_click': ['maxCPA1PerClick', 'Max. CPA (1-per-click)'],
    'max_cpc': ['maxCPC', 'Max. CPC'],
    'max_cpc_source': ['maxCPCSource', 'Max CPC source'],
    'max_cpm': ['maxCPM', 'Max. CPM'],
    'max_cpm_source': ['maxCPMSource', 'Max CPM Source'],
    'max_cpp': ['maxCPP', 'Max. CPP'],
    'member_count': ['memberCount', 'Member Count'],
    'metro_area': ['metroArea', 'Metro area'],
    'month': ['month', 'Month'],
    'month_of_year': ['monthOfYear', 'Month of Year'],
    'most_specific_location': ['mostSpecificLocation',
                                'Most specific location'],
    'negative_keyword': ['negativeKeyword', 'Negative keyword'],
    'network': ['network', 'Network'],
    'network_with_search_partners': ['networkWithSearchPartners',
                                      'Network (with search partners)'],
    'page': ['page', 'Page'],
    'phone_bid_type': ['phoneBidType', 'Phone bid type'],
    'phone_calls': ['phoneCalls', 'Phone calls'],
    'phone_cost': ['phoneCost', 'Phone cost'],
    'phone_impressions': ['phoneImpressions', 'Phone impressions'],
    'placement': ['placement', 'Placement'],
    'placement_state': ['placementState', 'Placement state'],
    'position_ace_indicator': ['positionACEIndicator',
                                'Position ACE indicator'],
    'ptr': ['ptr', 'PTR'],
    'quality_score': ['qualityScore', 'Quality score'],
    'quarter': ['quarter', 'Quarter'],
    'reference_count': ['referenceCount', 'Reference Count'],
    'region': ['region', 'Region'],
    'relative_ctr': ['relativeCTR', 'Relative CTR'],
    'search_exact_match_is': ['searchExactMatchIS', 'Search Exact match IS'],
    'search_impr_share': ['searchImprShare', 'Search Impr. share'],
    'search_lost_is_budget': ['searchLostISBudget', 'Search Lost IS (budget)'],
    'search_lost_is_rank': ['searchLostISRank', 'Search Lost IS (rank)'],
    'search_term': ['searchTerm', 'Search term'],
    'second_level_sub_categories': ['secondLevelSubCategories',
                                     'Second level sub-categories'],
    'shared_set_id': ['sharedSetID', 'Shared Set ID'],
    'shared_set_name': ['sharedSetName', 'Shared Set Name'],
    'shared_set_type': ['sharedSetType', 'Shared Set Type'],
    'start_time': ['startTime', 'Start time'],
    'state': ['state', 'State'],
    'status': ['status', 'Status'],
    'targeting_mode': ['targetingMode', 'Targeting Mode'],
    'this_extension_vs_other': ['thisExtensionVsOther',
                                 'This extension vs. Other'],
    'time_zone': ['timeZone', 'Time zone'],
    'top_level_categories': ['topLevelCategories', 'Top level categories'],
    'top_of_page_cpc': ['topOfPageCPC', 'Top of page CPC'],
    'top_vs_side': ['topVsSide', 'Top vs. side'],
    'topic': ['topic', 'Topic'],
    'topic_state': ['topicState', 'Topic state'],
    'total_conv_value': ['totalConvValue', 'Total conv. value'],
    'total_cost': ['totalCost', 'Total cost'],
    'unique_users': ['uniqueUsers', 'Unique Users'],
    'url': ['url', 'URL'],
    'value_conv1_per_click': ['valueConv1PerClick',
                               'Value / conv. (1-per-click)'],
    'value_conv_many_per_click': ['valueConvManyPerClick',
                                   'Value / conv. (many-per-click)'],
    'view_through_conv': ['viewThroughConv', 'View-through conv.'],
    'view_through_conv_ace_indicator': ['viewThroughConvACEIndicator',
                                         'View-through conv. ACE indicator'],
    'week': ['week', 'Week'],
    'year': ['year', 'Year']}


import logging
from ut.util.importing import get_environment_variable
from adwords.awq import AWQ
from adwords.connection import Connection
from adwords.gmoney import GMoney
from adwords.job_checker import JobChecker
from adwords.job_results import JobResults
from serialize.khan_logger import KhanLogger
from adwords.mutations import Mutations
from adwords.operations import Operations


class KeywordOperationsBase(object):
    def __init__(self, account_id='7998744469', store=None, min_money=None, max_money=None, logging_level=logging.DEBUG, chk_size=5000):
        """
        Pass in min and max (in Euros, NOT micros) based bid if you want to override the GMoney defaults
        store: storage that acts like an HDF5 store
        """
        self.conn = Connection(password=get_environment_variable('VEN_ADWORDS_PASSWORD'),
                               developer_token=get_environment_variable('VEN_ADWORDS_TOKEN'),
                               account_id=account_id)
        self.awq = AWQ(self.conn)
        self.gmoney = GMoney(min_money=min_money, max_money=max_money)
        self.ops = Operations(self.gmoney)
        self.mutations = Mutations(self.conn)

        self.chk_size = chk_size

        self.store = store

        self.logger = KhanLogger(level=logging_level, origin=self.__class__.__name__)

        # convenience properties to make it easier to call methods in succession without tracking job ids
        self.job_ids = None
        self.job_ids_completed = None
        self.job_ids_failed = None

    def _contains(self, list1, list2):
        """
        Returns true if list1 contains list2, otherwise returns false
        """
        if set(list1) - set(list2):
            return False
        else:
            return True

    # TODO - create another method that allows pinging to see if all done, so can cycle through accounts
    def get_completed_and_failed_jobs(self, job_ids=None):
        """
        Don't pass in job ids if you want to check on all job ids resulting from the last set of operations
        """
        job_ids = job_ids or self.job_ids
        jc = JobChecker(adwords_connection=self.conn, job_ids=job_ids)

        jc.poll_until_complete()
        self.job_ids_completed =jc.job_ids_completed
        self.job_ids_failed = jc.job_ids_failed

        #if self.job_ids_failed:
        #    self.logger.structured_log(level=logging.WARN, origin='keyword operations', msg="Failed jobs: {}".format(self.job_ids_failed))

        return self.job_ids_completed, self.job_ids_failed

    def get_errors(self, job_ids=None):
        """
        Don't pass in job_ids list if you want to get errors for all completed jobs
        """
        job_ids = job_ids or self.job_ids_completed
        jr = JobResults(adwords_connection=self.conn, job_ids=job_ids, logging_level=logging.ERROR)
        errors_by_job = jr.get_errors()
        for job_id in list(errors_by_job.keys()):
            locations = [t[0] for t in errors_by_job[job_id]]
            errors = [t[1] for t in errors_by_job[job_id]]
            df = self.store['jobs/_' + job_id.replace('-', 'N')]
            error_df = df.iloc[locations]
            error_df['error']=errors
            self.store.put('failed_jobs/_' + job_id.replace('-', 'N'), error_df)
        # return all of the job_ids that had errors
        return list(errors_by_job.keys())


__author__ = 'thorwhalen'
"""
Includes various adwords elements diagnosis functions
"""

#from daf.manip import lower_series

# from numpy.lib import arraysetops
# import pandas as pd

def ad_group_ids_are_unique(df):
    """
    This function returns True iff ad_group_ids are unique (only show up once in the rows of df)
    """
    return len(df[['ad_group_id']].drop_duplicates()) == len(df)

def ad_group_id_does_not_need_campaign_id(df):
    """
    This function returns True iff ad_group_ids are unique globally (that is, without the need of campaign_id)
    """
    return len(df[['ad_group_id']].drop_duplicates()) == len(df[['ad_group_id', 'campaign_id']].drop_duplicates())

def keyword_ids_are_unique(df):
    """
    This function returns True iff keyword_ids are unique (only show up once in the rows of df)
    """
    return len(df[['keyword_id']].drop_duplicates()) == len(df)

def keyword_id_does_not_need_ad_group_and_campaign_ids(df):
    """
    This function returns True iff keyword_ids are unique globally (that is, without the need of ag and cp ids)
    """
    return len(df[['keyword_id']].drop_duplicates())\
           == len(df[['keyword_id', 'ad_group_id', 'campaign_id']].drop_duplicates())

# def get_non_low_str_of_col(d,col):
#     return d[d[col]!=lower_series(d[col])]






__author__ = 'thor'

import ut.daf.ch as daf_ch
import ut.pdict.to as pdict_to

awe_col_synonyms = {
    'Campaign': ['campaign'],
    'Ad Group': ['ad_group', 'ad_group_name'],
    'Keyword': ['keyword', 'keyword_text'],
    'Max. CPC': ['max_cpc'],
    'match type': ['match_type'],
    'Destination URL': ['destination_url']
}

awe_col_replacer = pdict_to.word_replacer(pdict_to.inverse_one_to_many(awe_col_synonyms))


def mk_awe_cols(df):
    old_cols = df.columns
    new_cols = list(map(awe_col_replacer, old_cols))
    return daf_ch.ch_col_names(df, new_cols, old_cols)

